@@ -5,6 +5,8 @@ import json
 import boto3
 from mesh_client import MeshClient
 
+REGION_NAME = os.environ.get("AWS_REGION", "eu-west-2")
+
 
 class SingletonCheckFailure(Exception):
     """Singleton check failed"""
@@ -94,11 +96,11 @@ class MeshCommon:
         }
 
     @staticmethod
-    def get_ssm_params(path, recursive=False, decryption=True):
-        """Get parameters from SSM param store and return as simple dict"""
-        # TODO region name fix
-        ssm_client = boto3.client("ssm", region_name="eu-west-2")
-
+    def get_params(path, recursive=False, decryption=True):
+        """
+        Get parameters from SSM and secrets manager
+        """
+        ssm_client = boto3.client("ssm", region_name=REGION_NAME)
         params_result = ssm_client.get_parameters_by_path(
             Path=path,
             Recursive=recursive,
@@ -111,6 +113,18 @@ class MeshCommon:
             if name:
                 var_name = os.path.basename(name)
                 new_params_dict[var_name] = entry.get("Value", None)
+        if os.environ.get("use_secrets_manager") == "true":
+            secrets_client = boto3.client("secretsmanager", region_name=REGION_NAME)
+            all_secrets_dict = secrets_client.list_secrets()
+            all_secrets_list = all_secrets_dict["SecretList"]
+            for secret in all_secrets_list:
+                name = secret["Name"]
+                if name.startswith(path):
+                    secret_value = secrets_client.get_secret_value(SecretId=name)[
+                        "SecretString"
+                    ]
+                    var_name = os.path.basename(name)
+                    new_params_dict[var_name] = secret_value
         return new_params_dict
 
 
